#include <iostream>
#include <cstdio>
#include <string>
#include <vector>
//#include <chrono> // for timing
#include <limits> // numeric_limits
#include <memory> // for smart pointers
#include <ctime>

// COIN-OR
#include <CbcBranchDefaultDecision.hpp>
#include <CbcCompareObjective.hpp>
#include <CbcEventHandler.hpp>
#include <CbcModel.hpp>
#include <CbcTree.hpp>
#include <CglGMI.hpp>
#include <OsiClpSolverInterface.hpp>

#include "OsiChooseStrongCustom.hpp"

using SolverInterface = OsiClpSolverInterface;

// Main file variables
OsiSolverInterface *solver;
std::string dir = "", f_name = "", filename_stub = "", instname = "", in_file_ext = "";
std::time_t start_time_t, end_time_t;
char start_time_string[25];

#define macro_to_string(s) #s
#define x_macro_to_string(s) macro_to_string(s)

#ifdef MY_CBC_VERSION
const std::string CBC_VERSION_STRING = x_macro_to_string(MY_CBC_VERSION);
#endif
#ifdef MY_CLP_VERSION
const std::string CLP_VERSION_STRING = x_macro_to_string(MY_CLP_VERSION);
#endif
#ifdef USE_GUROBI
#include <gurobi_c++.h>
const std::string GUROBI_VERSION_STRING = std::to_string(GRB_VERSION_MAJOR) + "." + std::to_string(GRB_VERSION_MINOR) + std::to_string(GRB_VERSION_TECHNICAL);
#endif // USE_GUROBI

// Catch abort signal if it ever gets sent
/**
 * Catch a signal. You can output debugging info.
 * If you return from this function, and it was called
 * because abort() was called, your program will exit or crash anyway
 * (with a dialog box on Windows).
 */
#include <csignal>
void signal_handler_with_error_msg(int signal_number) {
  fprintf(stderr, "Abort or seg fault message received. Signal number: %d.\n", signal_number);
  exit(1);
} /* signal_handler_with_error_msg */

int startUp(int argc, char** argv);
int parseFilename(std::string& dir, std::string& instname, std::string& in_file_ext, const std::string& fullfilename);
void initializeSolver(OsiSolverInterface* &solver);
int wrapUp(int retCode);

/// An event handler that records the time that a better solution is found
class CustomHandler : public CbcEventHandler {
 public:
  CustomHandler() { };

  virtual CbcEventHandler* clone() const {
    return new CustomHandler(*this);
  }

  virtual CbcAction event(CbcEventHandler::CbcEvent whichEvent) {
    if (whichEvent == CbcEventHandler::treeStatus) {
      const int numNodesOnTree_ = model_->tree()->size();
      int numLeafNodes_ = 0;
      for (int i = 0; i < numNodesOnTree_; i++) {
        numLeafNodes_ += model_->tree()->nodePointer(i)->nodeInfo()->numberBranchesLeft();
      }
      printf("\n## Number nodes on tree: %d. Number leaf nodes: %d. Depth: %d. Node count: %d. ##\n",
          numNodesOnTree_, numLeafNodes_, model_->currentDepth(), model_->getNodeCount2());
      const int tmp = 1;
    }
    return CbcEventHandler::noAction;
  }
};

/****************** MAIN FUNCTION **********************/
int main(int argc, char** argv) {
  // Do this early in your program's initialization
  std::signal(SIGABRT, signal_handler_with_error_msg);
  std::signal(SIGSEGV, signal_handler_with_error_msg);

  // Process args and print welcome message
  int status = startUp(argc, argv);
  if (status) { return status; }

  // Set up solver and get initial solution
  initializeSolver(solver);
  solver->initialSolve();

  {
    // Create clone of solver to be used within b&b
    SolverInterface* BBSolver;
    BBSolver = dynamic_cast<SolverInterface*>(solver->clone());
    BBSolver->setHintParam(OsiDoPresolveInInitial, false);
    BBSolver->setHintParam(OsiDoPresolveInResolve, false);
    BBSolver->setIntParam(OsiMaxNumIterationHotStart, std::numeric_limits<int>::max());
    BBSolver->setSpecialOptions(16); // use standard strong branching
    // Do not switch from dual to primal, or something to this effect
    // This allows infeasible branches to be fixed during strong branching
    BBSolver->getModelPtr()->setMoreSpecialOptions(BBSolver->getModelPtr()->moreSpecialOptions() + 256);

    CbcModel* cbc_model = new CbcModel;
    cbc_model->swapSolver(BBSolver);
    cbc_model->setModelOwnsSolver(true);
    cbc_model->setLogLevel(3);

    // Set Cbc settings
    const int max_time = 120;
    const int num_strong = 5; //std::numeric_limits<int>::max();
    const int num_before_trusted = std::numeric_limits<int>::max(); // 10;
    cbc_model->setPrintFrequency(0); // to reach tree status correctly in latest Cbc
#ifdef CBC_VERSION_TRUNK
    cbc_model->setSecsPrintFrequency(-1); // to reach tree status correctly in latest Cbc
#endif
    cbc_model->solver()->setIntParam(OsiMaxNumIterationHotStart, 100);
    cbc_model->setTypePresolve(0);
    cbc_model->setMaximumSeconds(max_time);
    cbc_model->setMaximumCutPassesAtRoot(0);
    cbc_model->setMaximumCutPasses(0);
    cbc_model->setWhenCuts(0);
    if (num_strong >= 0) {
      // Maximum number of strong branching candidates to consider each time
      cbc_model->setNumberStrong(num_strong);
    }
    if (num_before_trusted >= 0) {
      // # before switching to pseudocosts, I think; 0 disables dynamic strong branching, doesn't work well
      cbc_model->setNumberBeforeTrust(num_before_trusted);
    }

    CbcBranchDecision* branch = new CbcBranchDefaultDecision();
    if (branch) { cbc_model->setBranchingMethod(*branch); }

    CbcCompareBase* compare = new CbcCompareObjective();
    if (compare) { cbc_model->setNodeComparison(compare); }

    CustomHandler* eventHandler = new CustomHandler;
    if (eventHandler) { cbc_model->passInEventHandler(eventHandler); }

    if (branch) {
      OsiChooseStrongCustom choose;
      const int choose_strategy = 0;
      if (num_strong >= 0) {
        choose.setNumberStrong(num_strong);
      }
      if (num_before_trusted >= 0) {
        choose.setNumberBeforeTrusted(num_before_trusted);
      }
      choose.setMethod(choose_strategy);
      branch->setChooseMethod(choose);
      // From CbcModel::convertToDynamic, we see that the branching decision may be ignored without a choose method
      if ((branch->whichMethod()&1) == 0 && !branch->chooseMethod()) {
        OsiChooseStrong choose;
        choose.setNumberStrong(5);
        choose.setNumberBeforeTrusted(0);
        branch->setChooseMethod(choose);
      }
      cbc_model->setBranchingMethod(*branch);
    } // check that branch is not NULL

    // Free memory after setting Cbc parameters
    if (eventHandler) { delete eventHandler; }
    if (branch) { delete branch; }
    if (compare) { delete compare; }

    // Do branch-and-bound
#ifdef TRACE
    cbc_model->branchAndBound(3);
#else
    cbc_model->branchAndBound(0);
#endif

    if (BBSolver && !cbc_model->modelOwnsSolver()) { delete BBSolver; }
    if (cbc_model) { delete cbc_model; }
  }
  return wrapUp(0);
} /* main */

/**
 * Call this early to print welcome message, etc.
 */
int startUp(int argc, char** argv) {
  int status = 0;

  // Input handling
  printf("## bb_example ##\n");
#ifdef MY_CBC_VERSION
  printf("## Cbc Version %s ##\n", CBC_VERSION_STRING.substr(0,8).c_str());
#endif
#ifdef MY_CLP_VERSION
  printf("## Clp Version %s ##\n", CLP_VERSION_STRING.substr(0,8).c_str());
#endif
#ifdef USE_GUROBI
  printf("## Gurobi Version %s ##\n", GUROBI_VERSION_STRING.c_str());
#endif
  printf("# Aleksandr M. Kazachkov\n");
  for (int i = 0; i < argc; i++) {
    std::cout << argv[i] << " ";
  }
  std::cout << std::endl;

  time(&start_time_t);
  struct tm* start_timeinfo = localtime(&start_time_t);
  snprintf(start_time_string, sizeof(start_time_string) / sizeof(char), "%s", asctime(start_timeinfo));
  printf("Start time: %s\n", start_time_string);

  if (argc < 2) {
    printf("Name of the input file (.lp or .mps):\n");
    std::getline(std::cin, f_name);
  } else {
    f_name = argv[1];
  }

  if (parseFilename(dir, instname, in_file_ext, f_name) != 0) {
    fprintf(stderr,
        "Unable to parse filename: %s. Found: dir=\"%s\", instname=\"%s\",ext=\"%s\".\n",
        f_name.c_str(), dir.c_str(),
        instname.c_str(), in_file_ext.c_str());
    exit(1);
  }
  filename_stub = dir + "/" + instname;

  return status;
} /* startUp */

int wrapUp(int retCode /*= 0*/) {
  time(&end_time_t);
  struct tm* end_timeinfo = localtime(&end_time_t);
  char end_time_string[25];
  snprintf(end_time_string, sizeof(end_time_string) / sizeof(char), "%s", asctime(end_timeinfo));

  printf("\n## Exiting bb_example. ##\n");
#ifdef MY_CBC_VERSION
  printf("Cbc Version: %s\n", CBC_VERSION_STRING.substr(0,8).c_str());
#endif
#ifdef MY_CLP_VERSION
  printf("Clp Version: %s\n", CLP_VERSION_STRING.substr(0,8).c_str());
#endif
#ifdef USE_GUROBI
  printf("Gurobi Version: %s\n", GUROBI_VERSION_STRING.c_str());
#endif
  printf("Instance: %s\n", instname.c_str());
  printf("Start time: %s\n", start_time_string);
  printf("End time: %s\n", end_time_string);
  printf("Elapsed time: %.f seconds\n", difftime(end_time_t, start_time_t));

  if (solver) {
    delete solver;
  }
  return retCode;
} /* wrapUp */

void initializeSolver(OsiSolverInterface* &solver) {
  // Generate cuts
  solver = new SolverInterface;

  int status = 0;
  if (in_file_ext.compare("lp") == 0) {
#ifdef TRACE
    printf("\n## Reading LP file. ##\n");
#endif
    status = solver->readLp(f_name.c_str());
  } else {
    if (in_file_ext.compare("mps") == 0) {
#ifdef TRACE
      printf("\n## Reading MPS file. ##\n");
#endif
      status = solver->readMps(f_name.c_str());
    } else {
      try {
#ifdef TRACE
        printf("\n## Reading MPS file. ##\n");
#endif
        status = solver->readMps(f_name.c_str());
      } catch (std::exception& e) {
        fprintf(stderr, "Unrecognized extension: %s.\n",
            in_file_ext.c_str());
        exit(1);
      }
    }
  } // read file
  if (status < 0) {
    fprintf(stderr, "Unable to read in file %s.\n",
        f_name.c_str());
    exit(1);
  }

  // Make sure we are doing a minimization problem; this is just to make later
  // comparisons simpler (i.e., a higher LP obj after adding the cut is better).
  if (solver->getObjSense() < 1e-3) {
    printf(
        "\n## Detected maximization problem. Negating objective function to make it minimization. ##\n");
    solver->setObjSense(1.0);
    const double* obj = solver->getObjCoefficients();
    for (int col = 0; col < solver->getNumCols(); col++) {
      solver->setObjCoeff(col, -1. * obj[col]);
    }
    double objOffset = 0.;
    solver->getDblParam(OsiDblParam::OsiObjOffset, objOffset);
    if (objOffset != 0.) {
      solver->setDblParam(OsiDblParam::OsiObjOffset, -1. * objOffset);
    }
  }
} /* initializeSolver */

/** Separate filename into the directory, instance name, and extension */
int parseFilename(std::string& dir, std::string& instname, std::string& in_file_ext, const std::string& fullfilename) {
  // Get file name stub
  size_t found_dot = fullfilename.find_last_of(".");
  std::string filename = fullfilename.substr(0, found_dot);

  // Put string after last '.' into string in_file_ext
  if (found_dot >= fullfilename.length()) {
    fprintf(stderr, "Cannot find the file extension (no '.' in input file name: \"%s\").\n", fullfilename.c_str());
    return 1;
  }

  // Check if archived file
  in_file_ext = fullfilename.substr(found_dot + 1);
  if (in_file_ext.compare("gz") == 0 || in_file_ext.compare("bz2") == 0) {
    unsigned found_dot_tmp = filename.find_last_of('.');

    // Put string after last '.' into string in_file_ext
    if (found_dot_tmp < filename.length()) {
      in_file_ext = filename.substr(found_dot_tmp + 1);
      filename = filename.substr(0, found_dot_tmp);
    } else {
      fprintf(stderr,
          "Other than gz or bz2, cannot find the file extension (no '.' in input file name: \"%s\").\n",
          fullfilename.c_str());
      return 1;
    }
  }

  //  const std::string filestub = (slashindex != std::string::npos) ? fullfilename.substr(slashindex+1) : fullfilename;
  size_t slashindex = filename.find_last_of("/\\");
  dir = (slashindex != std::string::npos) ? filename.substr(0,slashindex) : ".";
  instname = (slashindex != std::string::npos) ? filename.substr(slashindex+1) : filename;
  return 0;
} /* parseFilename */
